# -*- coding: utf-8 -*-
# Copyright 2015-2016 Fabian Hofmann (FIAS), Jonas Hoersch (FIAS)

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, print_function
import logging
import os
import subprocess as sub
import shutil
import tempfile
import pandas as pd
import numpy as np
from .core import _package_data
logger = logging.getLogger(__name__)


def add_geoposition_for_duke(df):
    """
    Returns the same pandas.Dataframe with an additional column "Geoposition"
    which concats the latitude and longitude of the powerplant in a string

    """
    if not df.loc[:, ['lat', 'lon']].isnull().all().all():
        return df.assign(Geoposition = df[['lat', 'lon']].astype(str)
                         .apply(lambda s: ','.join(s), axis=1)
                         .replace('nan,nan', np.nan))
    else:
        return df.assign(Geoposition=np.nan)


def duke(datasets, labels=['one', 'two'], singlematch=False,
         showmatches=False, keepfiles=False, showoutput=False):
    """
    Run duke in different modes (Deduplication or Record Linkage Mode) to
    either locate duplicates in one database or find the similar entries in two
    different datasets. In RecordLinkagesMode (match two databases) please
    set singlematch=True and use best_matches() afterwards

    Parameters
    ----------

    datasets : pd.DataFrame or [pd.DataFrame]
        A single dataframe is run in deduplication mode, while multiple ones
        are linked
    labels : [str], default ['one', 'two']
        Labels for the linked dataframe
    singlematch: boolean, default False
        Only in Record Linkage Mode. Only report the best match for each entry
        of the first named dataset. This does not guarantee a unique match in
        the second named dataset.
    keepfiles : boolean, default False
        If true, do not delete temporary files
    """

    dedup = isinstance(datasets, pd.DataFrame)
    if dedup:
        # Deduplication mode
        duke_config = "Deleteduplicates.xml"
        datasets = [datasets]
    else:
        duke_config = "Comparison.xml"

    duke_bin_dir = _package_data('duke_binaries')

    os.environ['CLASSPATH'] = \
        os.pathsep.join([os.path.join(duke_bin_dir, r)
                        for r in os.listdir(duke_bin_dir)])
    tmpdir = tempfile.mkdtemp()

    try:
        shutil.copyfile(os.path.join(_package_data(duke_config)),
                        os.path.join(tmpdir, "config.xml"))

        logger.debug("Comparing files: %s", ", ".join(labels))

        for n, df in enumerate(datasets):
            df = add_geoposition_for_duke(df)
#            due to index unity (see https://github.com/larsga/Duke/issues/236)
            if n == 1:
                shift_by = (datasets[0].index.max()+1)
                df.index += shift_by
            df.to_csv(os.path.join(tmpdir, "file{}.csv".format(n+1)),
                      index_label='id')
            if n == 1:
                df.index -= shift_by

        args = ['java', '-Dfile.encoding=UTF-8', 'no.priv.garshol.duke.Duke',
                '--linkfile=linkfile.txt']
        if singlematch:
            args.append('--singlematch')
        if showmatches:
            args.append('--showmatches')
            stdout = sub.PIPE
        else:
            stdout = None
        args.append('config.xml')

        run = sub.Popen(args, stderr=sub.PIPE, cwd=tmpdir, stdout=stdout,
                        universal_newlines=True)
        _, stderr = run.communicate()

        if showmatches:
            print(_)

        logger.debug("Stderr: {}".format(stderr))
        if any(word in stderr.lower() for word in ['error', 'fehler']):
            raise RuntimeError("duke failed: {}".format(stderr))

        if dedup:
            return pd.read_csv(os.path.join(tmpdir, 'linkfile.txt'),
                               encoding='utf-8', usecols=[1, 2], names=labels)
        else:
            res = pd.read_csv(os.path.join(tmpdir, 'linkfile.txt'),
                             usecols=[1, 2, 3], names=labels + ['scores'])
            res.iloc[:, 1] -= shift_by
            return res

    finally:
        if keepfiles:
            logger.debug('Files of the duke run are kept in {}'.format(tmpdir))
        else:
            shutil.rmtree(tmpdir)
            logger.debug('Files of the duke run have been deleted in {}'
                         .format(tmpdir))



# -*- coding: utf-8 -*-
# Copyright 2016-2018 Fabian Hofmann (FIAS), Jonas Hoersch (KIT, IAI) and
# Fabian Gotzens (FZJ, IEK-STE)

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import math
import numpy as np
import pandas as pd
#import collections
import matplotlib.pyplot as plt
#import matplotlib.patches as mpatches
from matplotlib.patches import Circle, Ellipse
from matplotlib.legend_handler import HandlerPatch
from matplotlib import rcParams, cycler
from matplotlib.lines import Line2D
import seaborn as sns

from .core import get_config, get_obj_if_Acc
from .utils import lookup, set_uncommon_fueltypes_to_other, to_list_if_other

import logging
logger = logging.getLogger(__name__)


cartopy_present = True
try:
    import cartopy.crs as ccrs
    import cartopy
except:
    cartopy_present = False

if not cartopy_present:
    logger.warn('Cartopy not existent.')



def fueltype_stats(df):
    df = get_obj_if_Acc(df)
    stats = lookup(set_uncommon_fueltypes_to_other(df), by='Fueltype')
    plt.pie(stats, colors=stats.index.to_series()
                               .map(get_config()['fuel_to_color'])
                               .fillna('gray'),
            labels=stats.index, autopct='%1.1f%%')


def powerplant_map(df, scale=1e2, european_bounds=True, fillcontinents=False,
                   legendscale=1, resolution=True, **kwargs):

    df = get_obj_if_Acc(df)
    # TODO: add reference circle in legend
    figsize = kwargs.get('figsize', (11, 9))
    with sns.axes_style('darkgrid'):
        df = set_uncommon_fueltypes_to_other(df)
        shown_fueltypes = df.Fueltype.unique()
        df = df[df.lat.notnull()]
        sub_kw = {'projection': ccrs.PlateCarree()} if cartopy_present else {}
        fig, ax = plt.subplots(figsize=figsize, subplot_kw=sub_kw)

        ax.scatter(df.lon, df.lat, s=df.Capacity/scale,
                   c=df.Fueltype.map(get_config()['fuel_to_color']),
                   edgecolor='face', facecolor='face')

        legendcols = (pd.Series(get_config()['fuel_to_color'])
                        .reindex(shown_fueltypes))
        handles = sum(legendcols.apply(lambda x:
                      make_legend_circles_for([10.], scale=scale*legendscale,
                                              facecolor=x)).tolist(), [])
        ax.legend(handles, legendcols.index,
                   handler_map=make_handler_map_to_scale_circles_as_in(ax),
                   markerscale=1,
                   ncol=kwargs.get('ncol', 2),
                   loc=kwargs.get('loc', "upper left"),
                   frameon=True, fancybox=True, edgecolor='w',
                   facecolor='w', framealpha=1)

        ax.set_xlabel('')
        ax.set_ylabel('')
        if european_bounds:
            ax.set_xlim(-13, 34)
            ax.set_ylim(35, 72)
        draw_basemap(ax=ax, resolution=resolution,
                     fillcontinents=fillcontinents)
        ax.set_facecolor('w')

        fig.tight_layout(pad=0.5)
        if cartopy_present:
            ax.outline_patch.set_visible(False)
        return fig, ax

# This approach is an alternative to bar_comparison_countries_fueltypes()
def fueltype_and_country_totals_bar(dfs, keys=None, figsize=(18, 8)):
    dfs = get_obj_if_Acc(dfs)
    dfs = to_list_if_other(dfs)
    df = lookup(dfs, keys)
    countries = df.index.get_level_values('Country').unique()
    n = len(countries)
    subplots = gather_nrows_ncols(n)
    fig, ax = plt.subplots(*subplots, figsize=figsize)

    if sum(subplots) > 2:
        ax_iter = ax.flat
    else:
        ax_iter = np.array(ax).flat
    for country in countries:
        ax = next(ax_iter)
        df.loc[country].plot.bar(ax=ax, sharex=True, legend=None)
        ax.set_xlabel('')
        ax.ticklabel_format(axis='y', style='sci', scilimits=(-2, 2))
        ax.set_title(country)
    handles, labels = ax.get_legend_handles_labels()
    fig.tight_layout(pad=1)
    fig.legend(handles, labels, loc=9, ncol=2, bbox_to_anchor=(0.53, 0.99))
    fig.subplots_adjust(top=0.9)
    return fig, ax


def fueltype_totals_bar(dfs, keys=None, figsize=(7, 4), unit='GW',
                        last_as_marker=False, axes_style='whitegrid',
                        exclude=[], **kwargs):
    dfs = get_obj_if_Acc(dfs)
    dfs = to_list_if_other(dfs)
    dfs = [df[~df.Fueltype.isin(exclude)] for df in dfs]
    with sns.axes_style(axes_style):
        fig, ax = plt.subplots(1, 1, figsize=figsize)
        if last_as_marker:
            as_marker = dfs[-1]
            dfs = dfs[:-1]
            as_marker_key = keys[-1]
            keys = keys[:-1]
        fueltotals = lookup(dfs, keys=keys, by='Fueltype', unit=unit)
        fueltotals.plot(kind="bar", ax=ax, legend='reverse',
                        edgecolor='none', rot=90, **kwargs)
        if last_as_marker:
            fueltotals = lookup(as_marker, keys=as_marker_key, by='Fueltype',
                                unit=unit)
            fueltotals.plot(ax=ax, label=as_marker_key, markeredgecolor='none',
                            rot=90, marker='D', linestyle='None',
                            markerfacecolor='darkslategray', **kwargs)
        ax.legend(loc=0)
        ax.set_ylabel(r'Capacity [$%s$]' % unit)
        ax.xaxis.grid(False)
        fig.tight_layout(pad=0.5)
        return fig, ax


def country_totals_hbar(dfs, keys=None, exclude_fueltypes=['Solar', 'Wind'],
                        figsize=(7, 5), unit='GW', axes_style='whitegrid'):
    with sns.axes_style(axes_style):
        fig, ax = plt.subplots(1, 1, figsize=figsize)
        countrytotals = lookup(dfs, keys=keys, by='Country',
                               exclude=exclude_fueltypes, unit=unit)
        countrytotals[::-1][1:].plot(kind="barh", ax=ax, legend='reverse',
                                     edgecolor='none')
        ax.set_xlabel('Capacity [%s]' % unit)
        ax.yaxis.grid(False)
        ax.set_ylabel('')
        fig.tight_layout(pad=0.5)
        return fig, ax


def factor_comparison(dfs, keys=None, figsize=(12, 9)):
    with sns.axes_style('whitegrid'):
        dfs = [set_uncommon_fueltypes_to_other(df) for df in dfs]
        compare = lookup(dfs, keys=keys, exclude=['Solar', 'Wind']).fillna(0.)
        compare = compare.append(
                pd.concat([compare.groupby(level='Country').sum()],
                          keys=['Total']).swaplevel()).sort_index()/1000
        n_countries, n_fueltypes = compare.index.levshape
        c = [get_config()['fuel_to_color'][i] for i in compare.index.levels[1]]
        rcParams["axes.prop_cycle"] = cycler(color=c)

        # where both are zero,
        compare[compare.sum(1) < 0.5] = np.nan

        fig, ax = plt.subplots(1, 1, figsize=figsize)
        compare = (compare.unstack('Country').swaplevel(axis=1).sort_index(axis=1)
                   .reindex(columns=keys, level=1))
        compare.T.plot(ax=ax, markevery=(0, 2),
                       style='o', markersize=5)
        compare.T.plot(ax=ax, markevery=(1, 2),
                       style='s', legend=None, markersize=4.5)

        lgd = ax.get_legend_handles_labels()

        for i, j in enumerate(compare.columns.levels[0]):
            ax.plot(np.array([0, 1]) + (2*i), compare[j].T)

        indexhandles = [Line2D([0.4, 0.6], [0.4, 0.6], marker=m, linewidth=0.,
                        markersize=msize, color='w', markeredgecolor='k',
                        markeredgewidth=0.5)
                        for m, msize in [['o', 5.], ['s', 4.5]]]
        ax.add_artist(ax.legend(handles=indexhandles, labels=keys))
        ax.legend(handles=lgd[0][:len(c)], labels=lgd[1][:len(c)],
                  title=False, loc=2)

        ax.set_xlim(-1, n_countries*2 + 1)
        ax.xaxis.grid(False)
        ax.set_xticks(np.linspace(0.5, n_countries*2 - 1.5, n_countries))
        ax.set_xticklabels(compare.columns.levels[0].values, rotation=90)
        ax.set_xlabel('')
        ax.set_ylabel('Capacity [GW]')
        fig.tight_layout(pad=0.5)
        return fig, ax


def boxplot_gross_to_net(axes_style='darkgrid', **kwargs):
    """
    """
    from .heuristics import gross_to_net_factors as gtn
    with sns.axes_style(axes_style):
        df = (gtn(return_entire_data=True)
              .loc[lambda df: df.energy_source_level_2 != 'Hydro'])
        df.loc[:, 'FuelTech'] = (df.energy_source_level_2
                                 + '\n(' + df.technology + ')')
        df = df.groupby('FuelTech').filter(lambda x: len(x) >= 10)
        dfg = df.groupby('FuelTech')
        # plot
        fig, ax = plt.subplots(**kwargs)
        df.boxplot(ax=ax, column='ratio', by='FuelTech', rot=90,
                   showmeans=True)
        ax.title.set_visible(False)
        ax.set_ylabel('Ratio of gross/net [$-$]')
        ax.xaxis.label.set_visible(False)
        ax.set_ylabel('Net / Gross')
        ax2 = ax.twiny()
        ax2.set_xlim(ax.get_xlim())
        ax2.set_xticks([i+1 for i in range(len(dfg))])
        ax2.set_xticklabels(['$n$=%d' % (len(v)) for k, v in dfg])
        fig.suptitle('')
        return fig, ax


def boxplot_matchcount(df):
    """
    Makes a boxplot for the capacities grouped by the number of matches.
    Attention: Currently only works for the full dataset with original
    names as the last columns.
    """
    # Mend needed data
    df.loc[:, 'Matches'] = df.projectID.apply(len)

    # Plot
    fig, ax = plt.subplots(figsize=(8, 4.5))
    df.boxplot(ax=ax, column='Capacity', by='Matches', showmeans=True)
    ax.title.set_visible(False)
    ax.set_xlabel('Number of datasets participating in match [$-$]')
    ax.set_ylabel('Capacity [$MW$]')
    fig.suptitle('')
    return fig


def make_handler_map_to_scale_circles_as_in(ax,
                                            dont_resize_actively=False):
    fig = ax.get_figure()

    def axes2pt():
        return np.diff(
                ax.transData.transform([(0, 0), (1, 1)]),
                axis=0)[0] * (72./fig.dpi)
    ellipses = []
    if not dont_resize_actively:
        def update_width_height(event):
            dist = axes2pt()
            for e, radius in ellipses:
                e.width, e.height = 2. * radius * dist
        fig.canvas.mpl_connect('resize_event', update_width_height)
        ax.callbacks.connect('xlim_changed', update_width_height)
        ax.callbacks.connect('ylim_changed', update_width_height)

    def legend_circle_handler(legend, orig_handle, xdescent, ydescent,
                              width, height, fontsize):
        w, h = 2. * orig_handle.get_radius() * axes2pt()
        e = Ellipse(xy=(0.5*width - 0.5*xdescent,
                        0.5*height - 0.5*ydescent),
                    width=w, height=w)
        ellipses.append((e, orig_handle.get_radius()))
        return e
    return {Circle: HandlerPatch(patch_func=legend_circle_handler)}


def make_legend_circles_for(sizes, scale=1.0, **kw):
    return [Circle((0, 0), radius=(s/scale)**0.5, **kw) for s in sizes]


def draw_basemap(resolution=True, ax=None, country_linewidth=0.3,
                 coast_linewidth=0.4, zorder=None, fillcontinents=True,
                 **kwds):

    if cartopy_present:
        if ax is None:
            ax = plt.gca(projection=ccrs.PlateCarree())
        resolution = '50m' if isinstance(resolution, bool) else resolution
        assert resolution in ['10m', '50m', '110m'], ('Resolution has to be '
                             "one of '10m', '50m', '110m'")
        ax.set_extent(ax.get_xlim() + ax.get_ylim(),
                      crs=ccrs.PlateCarree())
        ax.coastlines(linewidth=0.4, zorder=-1, resolution=resolution)
        border = cartopy.feature.BORDERS.with_scale(resolution)
        ax.add_feature(border, linewidth=0.3)
        ax.outline_patch.set_visible(False)
        if fillcontinents:
            land = cartopy.feature.LAND.with_scale(resolution)
            ax.add_feature(land, facecolor='lavender', alpha=0.25)


def gather_nrows_ncols(x, orientation='landscape'):
    """
    Derives [nrows, ncols] based on x plots, so that a subplot looks nicely.

    Parameters
    ----------
    x : int, Number of subplots between [0, 42]
    """
    def calc(n, m):
        if n <= 0:
            n = 1
        if m <= 0:
            m = 1
        while (n*m < x):
            m += 1
        return n, m

    if not isinstance(x, int):
        raise ValueError('An integer needs to be passed to this function.')
    elif x <= 0:
        raise ValueError('The given number of subplots is less or equal zero.')
    elif x > 42:
        raise ValueError("Are you sure that you want to put more than 42 "
                         "subplots in one diagram?\n You better don't, it "
                         "looks squeezed. Otherwise adapt the code.")
    k = math.sqrt(x)
    if k.is_integer():
        return [int(k), int(k), 0]  # Square format!
    else:
        k = int(math.floor(k))
        # Solution 1
        n, m = calc(k, k+1)
        sol1 = {'n': n, 'm': m, 'dif': (m*n) - x}
        # Solution 2:
        n, m = calc(k-1, k+1)
        sol2 = {'n': n, 'm': m, 'dif': (m*n) - x}
        if (((sol1['dif'] <= sol2['dif']) & (sol1['n'] >= 2)) |
                (x in [7, 13, 14])):
            n, m = [sol1['n'], sol1['m']]
        else:
            n, m = [sol2['n'], sol2['m']]
        remainder = m*n - x
        if orientation == 'landscape':
            return n, m, remainder
        elif orientation == 'portrait':
            return m, n, remainder
        else:
            raise ValueError('Wrong `orientation` given!')



#%% extra data needed, plots for publication
#
#
#def Show_all_plots():
#    comparison_single_matched_bar()
#    comparison_1dim(by='Country')
#    comparison_1dim(by='Fueltype')
#    return
#
#
#def comparison_countries_fueltypes_bar(
#        dfs=None, include_WEPP=True, include_VRE=False, exclude=None,
#        show_indicators=True, year=2015, **kwargs):
#    """
#    Plots per country an analysis, how the given datasets differ by fueltype.
#
#    Parameters
#    ----------
#    dfs : dict
#        keys : labels
#        values : pandas.Dataframe containing the data to be plotted
#    include_WEPP : bool
#        Switch to include WEPP-based data
#    include_VRE : bool
#        Switch to include VRE data
#    show_indicators : bool
#        Switch whether to calculate a coverage between to datasets
#    year : int
#        Only plot units with a commissioning year smaller or equal this value
#    """
#    threshold = kwargs.get('threshold', -1)
#    ylabel = kwargs.get('ylabel', u'Capacity [$GW$]')
#    mode = kwargs.get('mode', 'screen')  # valid: ['screen', 'print']
#    if mode == 'screen':
#        figsize = (27, 15)  # Ratio close to FullHD Resolution
#        orientation = 'landscape'
#        bottom_space = 0.14
#    elif mode == 'print':
#        figsize = (16.54, 23.38)  # For Din (A2, A3, A4, etc.) paper print
#        orientation = 'portrait'
#        bottom_space = 0.07
#    else:
#        raise ValueError('Wrong print mode given!')
#
#    countries = set()
#    if dfs is None:
#        red_w_wepp, red_wo_wepp, wepp, statistics = (
#                gather_comparison_data(include_WEPP=include_WEPP,
#                                       include_VRE=include_VRE, year=year))
#        if include_WEPP:
#            stats = lookup([red_w_wepp, red_wo_wepp, wepp, statistics],
#                           keys=['Matched dataset w/ WEPP',
#                                 'Matched dataset w/o WEPP',
#                                 'WEPP only', 'Statistics ENTSO-E SO&AF'],
#                           by='Country, Fueltype', exclude=exclude)/1000
#            set.update(countries, set(red_w_wepp.Country),
#                       set(red_wo_wepp.Country), set(wepp.Country),
#                       set(statistics.Country))
#        else:
#            stats = lookup([red_wo_wepp, statistics],
#                           keys=['Matched dataset w/o WEPP',
#                                 'Statistics ENTSO-E SO&AF'],
#                           by='Country, Fueltype', exclude=exclude)/1000
#            set.update(countries, set(red_wo_wepp.Country),
#                       set(statistics.Country))
#    else:
#        stats = lookup(dfs.values(), keys=dfs.keys(), by='Country, Fueltype',
#                       exclude=exclude)/1000
#        stats.sort_index(axis=1, inplace=True)
#        for k, v in dfs.items():
#            set.update(countries, set(v.Country))
#    # Filter stats
#    stats = (stats.replace({0.0: np.nan})     # Do this in order to show only
#                  .dropna(axis=0, how='all')  # relevant fueltypes for each
#                  .fillna(0.0))               # country (if all zero->drop!).
#
#    if (show_indicators or threshold >= 0.) and len(stats.columns) < 2:
#        logger.warn('At least two objects for comparison needed when using '
#                    '`show_indicators` or `threshold`. Arguments ignored.')
#        show_indicators = False
#        threshold = -1
#
#    # Presettings for the plots
#    font = {'size': 12}
#    plt.rc('font', **font)
#    # Loop through countries.
#    nrows, ncols, rem = gather_nrows_ncols(len(countries),
#                                           orientation=orientation)
#    i, j = [0, 0]
#    labels_mpatches = collections.OrderedDict()
#    fig, ax = plt.subplots(nrows=nrows, ncols=ncols, sharex=False,
#                           sharey=False, squeeze=False, figsize=figsize)
#    for country in sorted(countries):
#        if j == ncols:
#            i += 1
#            j = 0
#        # Perform the plot
#        stats.loc[country].plot.bar(ax=ax[i, j], stacked=False, legend=False,
#                                    colormap='jet')
#        if show_indicators or threshold >= 0.0:
#            # TODO: Assure that matched is always in 1st+stats in last column.
#            colm = stats.loc[country].columns[0]
#            cols = stats.loc[country].columns[-1]
#        if threshold >= 0.0:
#            ctry = stats.loc[country]
#            ctry.loc[:, 'ratio'] = abs(ctry[colm] - ctry[cols])/ctry[cols]
#            ctry.loc[:, 'delta'] = abs(ctry[colm] - ctry[cols])
#            ctry.loc[:, 'mean'] = ctry.loc[:, colm:cols].apply(np.mean, axis=1)
#            ctry = (ctry.reset_index(drop=True)
#                        .loc[lambda x: x['ratio'] >= threshold])
#            circles = [mpatches.Ellipse(xy=(float(x),  r['mean']),
#                                        height=(r['delta'])*2.5,
#                                        width=4./ctry.index.max(),
#                                        color='g', alpha=0.5)
#                       for x, r in ctry.iterrows()]
#            for c in circles:
#                ax[i, j].add_artist(c)
#        if show_indicators:
#            ctry = stats.loc[country]
#            r_sq = round(ctry.corr().iloc[0, 1]**2, 3)
#            cov = round(ctry[colm].sum() / ctry[cols].sum(), 3)
#            txt = AnchoredText(
#                    "\n" + r'$R^{2} = $%s' % r_sq + "\n"
#                    r'$\frac{\sum P_{match}}{\sum P_{stats}} = $%s' % cov,
#                    loc=get_config()['textbox_position'][country],
#                    prop={'size': 11})
#            txt.patch.set(boxstyle='round', alpha=0.5)
#            ax[i, j].add_artist(txt)
#        # Pass the legend information into the Ordered Dict
#        stats_handle, stats_labels = ax[i, j].get_legend_handles_labels()
#        for u, v in enumerate(stats_labels):
#            if v not in labels_mpatches:
#                labels_mpatches[v] = mpatches.Patch(
#                        color=stats_handle[u].patches[0].get_facecolor(),
#                        label=v)
#        if threshold >= 0.0:
#            label = 'Threshold (={}%) marker'.format(int(threshold*100.))
#            labels_mpatches[label] = mpatches.Patch(color='g', alpha=0.5,
#                                                    label=label)
#        # Format the subplots nicely
#        ax[i, j].set_facecolor('#d9d9d9')
#        ax[i, j].set_axisbelow(True)
#        ax[i, j].grid(color='white', linestyle='dotted')
#        ax[i, j].set_title(country)
#        ax[i, 0].set_ylabel(ylabel)
#        ax[i, j].xaxis.label.set_visible(False)
#        j += 1
#    # After the loop, do the rest of the layouting.
#    fig.tight_layout()
#    # Legend
#    fig.subplots_adjust(bottom=bottom_space)
#    labels_mpatches = collections.OrderedDict(sorted(
#            labels_mpatches.items()))
#    fig.legend(labels_mpatches.values(), labels_mpatches.keys(),
#               loc=8, ncol=len(labels_mpatches), facecolor='#d9d9d9')
#    return fig, ax
#
#
#def matched_fueltype_totals_bar(figsize=(9, 4), axes_style='whitegrid'):
#    from . import data
#    from .collection import Carma_ENTSOE_GEO_OPSD_WRI_matched_reduced
#    matched = set_uncommon_fueltypes_to_other(
#            Carma_ENTSOE_GEO_OPSD_WRI_matched_reduced())
#    matched.loc[matched.Fueltype == 'Waste', 'Fueltype'] = 'Other'
#    geo = set_uncommon_fueltypes_to_other(data.GEO())
#    carma = set_uncommon_fueltypes_to_other(data.CARMA())
#    wri = set_uncommon_fueltypes_to_other(data.WRI())
#    ese = set_uncommon_fueltypes_to_other(data.ESE())
#    entsoe = set_uncommon_fueltypes_to_other(data.Capacity_stats())
#    opsd = set_uncommon_fueltypes_to_other(data.OPSD())
#    entsoedata = set_uncommon_fueltypes_to_other(data.ENTSOE())
#
#    matched.Capacity = matched.Capacity/1000.
#    geo.Capacity = geo.Capacity/1000.
#    carma.Capacity = carma.Capacity/1000.
#    wri.Capacity = wri.Capacity/1000.
#    ese.Capacity = ese.Capacity/1000.
#    entsoe.Capacity = entsoe.Capacity/1000.
#    opsd.Capacity = opsd.Capacity/1000.
#    entsoedata.Capacity = entsoedata.Capacity/1000.
#
#    with sns.axes_style(axes_style):
#        fig, (ax1, ax2) = plt.subplots(1, 2, figsize=figsize, sharey=True)
#        databases = lookup(
#                [carma, entsoedata, ese, geo, opsd, wri],
#                keys=['CARMA', 'ENTSOE', 'ESE', 'GEO', 'OPSD', 'WRI'],
#                by='Fueltype')
#        databases = databases[databases.sum(1) > 5]
#        databases.plot(kind='bar', ax=ax1, edgecolor='none', rot=70)
#        datamatched = lookup(matched, by='Fueltype')
#        datamatched.index.name = ''
#        datamatched.name = 'Matched Database'
#        datamatched.plot(kind='bar', ax=ax2, color='steelblue',
#                         edgecolor='none', rot=70)
#        ax2.legend()
#        ax1.set_ylabel('Capacity [GW]')
#        ax1.xaxis.grid(False)
#        ax2.xaxis.grid(False)
#        fig.tight_layout(pad=0.5)
#        return fig, [ax1, ax2]
#
#
#
#def bar_decomissioning_curves(df=None, ylabel=None, title=None,
#                              legend_in_subplots=False):
#    """
#    Plots per country a decommissioning curve as a bar chart with
#    capacity on y-axis, period on x-axis and categorized by fueltype.
#    """
#    from .collection import \
#        Carma_ENTSOE_ESE_GEO_GPD_IWPDCY_OPSD_WEPP_matched_reduced_VRE
#    if ylabel is None:
#        ylabel = 'Capacity [GW]'
#    if df is None:
#        df = Carma_ENTSOE_ESE_GEO_GPD_IWPDCY_OPSD_WEPP_matched_reduced_VRE()
#        if df is None:
#            raise RuntimeError("The data to be plotted does not yet exist.")
#    df = df.copy()
#
#    df.loc[:, 'Life'] = df.Fueltype.map(get_config()['fuel_to_color'])
#
#    # Insert periodwise capacities
#    df.loc[:, 2015] = df.loc[:, 'Capacity']
#    for yr in range(2020, 2055, 5):
#        df.loc[yr <= (df.loc[:, 'YearCommissioned']
#                      + df.loc[:, 'Life']), yr] = df.loc[:, 'Capacity']
#        df.loc[:, yr].fillna(0., inplace=True)
#
#    # Presettings for the plots
#    font = {'size': 16}
#    plt.rc('font', **font)
#
#    nrows, ncols = gather_nrows_ncols(len(set(df.Country)))
#    fig, ax = plt.subplots(nrows=nrows, ncols=ncols, sharex=True, sharey=False,
#                           squeeze=False, figsize=(32/1.2, 18/1.2))
#    data_countries = df.groupby(['Country'])
#    i, j = [0, 0]
#    labels_mpatches = collections.OrderedDict()
#    for country in sorted(set(df.Country)):
#        if j == ncols:
#            i += 1
#            j = 0
#        cntry_grp = data_countries.get_group(country)
#        stats = pd.DataFrame(columns=range(2015, 2055, 5))
#        for yr in range(2015, 2055, 5):
#            k = cntry_grp.groupby(['Fueltype']).sum()/1000
#            stats.loc[:, yr] = k[yr]
#        colors = (stats.index.to_series()
#                  .map(get_config()['fuel_to_color'])
#                  .tolist())
#        stats.T.plot.bar(ax=ax[i, j], stacked=True, legend=False, color=colors)
#        # Pass the legend information into the Ordered Dict
#        if not legend_in_subplots:
#            stats_handle, stats_labels = ax[i, j].get_legend_handles_labels()
#            for u, v in enumerate(stats_labels):
#                if v not in labels_mpatches:
#                    labels_mpatches[v] = mpatches.Patch(color=colors[u],
#                                                        label=v)
#        else:
#            ax[i, j].legend(fontsize=9, loc='best')
#        # Format the subplots nicely
#        ax[i, j].set_facecolor('#d9d9d9')
#        ax[i, j].set_axisbelow(True)
#        ax[i, j].grid(color='white', linestyle='dotted')
#        ax[i, j].set_title(country)
#        ax[i, 0].set_ylabel(ylabel)
#        ax[-1, j].xaxis.label.set_visible(False)
#        j += 1
#    # After the loop, do the rest of the layouting.
#    fig.tight_layout()
#    if isinstance(title, str):
#        fig.suptitle(title, fontsize=24)
#        fig.subplots_adjust(top=0.93)
#    if not legend_in_subplots:
#        fig.subplots_adjust(bottom=0.08)
#        labels_mpatches = collections.OrderedDict(
#                sorted(labels_mpatches.items()))
#        fig.legend(labels_mpatches.values(), labels_mpatches.keys(),
#                   loc=8, ncol=len(labels_mpatches), facecolor='#d9d9d9')
#    return fig, ax
#
#
#def area_yearcommissioned(dfs, keys, figsize=(7, 5),
#                          ylabel='Capacity [$GW$]'):
#    """
#    Plots an area chart by commissioning year.
#
#    Parameters
#    ----------
#        dfs : list
#            containing pd.DataFrames to plot
#        keys : list
#            containing the names used as ax.title
#        figsize : tuple
#        ylabel : str
#    """
#    dfp = [df.pivot_table(values='Capacity', index='YearCommissioned',
#                          columns='Fueltype', aggfunc='sum')/1000
#           for df in dfs]
#    labels_mpatches = collections.OrderedDict()
#    fig, ax = plt.subplots(nrows=2, ncols=2, sharex=True, sharey=True,
#                           squeeze=False, figsize=figsize)
#    ax[0, 0].set_xlim((1900, 2016))
#    i, j = (0, 0)
#    for a, df in enumerate(dfp):
#        if j == 2:
#            i += 1
#            j = 0
#        colors = (df.columns.to_series()
#                  .map(get_config()['fuel_to_color'])
#                  .tolist())
#        df.plot.area(ax=ax[i, j], stacked=True, legend=False, color=colors,
#                     linewidth=0.0)
#        # Pass the legend information into the Ordered Dict
#        stats_handle, stats_labels = ax[i, j].get_legend_handles_labels()
#        for u, v in enumerate(stats_labels):
#            if v not in labels_mpatches:
#                labels_mpatches[v] = mpatches.Patch(color=colors[u], label=v)
#        ax[i, j].set_title(keys[a])
#        ax[i, j].set_facecolor('#d9d9d9')
#        ax[i, j].set_axisbelow(True)
#        ax[i, j].grid(color='white', linestyle='dotted')
#        ax[i, j].set_ylabel(ylabel)
#        ax[i, j].set_ylim((0, 31))
#        j += 1
#    fig.tight_layout()
#    # Legend
#    fig.subplots_adjust(bottom=0.25)
#    labels_mpatches = collections.OrderedDict(sorted(labels_mpatches.items()))
#    fig.legend(labels_mpatches.values(), labels_mpatches.keys(),
#               loc=8, ncol=len(labels_mpatches)/2, facecolor='#d9d9d9',
#               prop={'size': 10})
#    return fig, ax
#
#
## %% Plot utilities
#
#
#
#def comparison_single_matched_bar(df=None, include_WEPP=True, cleaned=True,
#                                  use_saved_aggregation=True, figsize=(9, 5),
#                                  exclude=['Geothermal', 'Solar', 'Wind'],
#                                  axes_style='whitegrid'):
#    """
#    Plots two bar charts for comparison
#    1.) Fueltypes on x-axis and capacity on y-axis, categorized by input db.
#    2.) Fueltypes on x-axis and capacity on y-axis, matched database.
#    """
#    from .cleaning import aggregate_units
#    from .data import CARMA, ENTSOE, ESE, GEO, OPSD, WEPP, WRI
#    from .collection import (
#            Carma_ENTSOE_ESE_GEO_GPD_IWPDCY_OPSD_WEPP_matched_reduced,
#            Carma_ENTSOE_ESE_GEO_GPD_IWPDCY_OPSD_matched_reduced)
#
#    if df is None:
#        if include_WEPP:
#            df = (Carma_ENTSOE_ESE_GEO_GPD_IWPDCY_OPSD_WEPP_matched_reduced()
#                  .loc[lambda x:  ~x.Fueltype.isin(exclude)])
#        else:
#            df = (Carma_ENTSOE_ESE_GEO_GPD_IWPDCY_OPSD_matched_reduced()
#                  .loc[lambda x:  ~x.Fueltype.isin(exclude)])
#
#    if cleaned:
#        carma = aggregate_units(CARMA(), dataset_name='CARMA',
#                                use_saved_aggregation=use_saved_aggregation)
#        entsoe = aggregate_units(ENTSOE(), dataset_name='ENTSOE',
#                                 use_saved_aggregation=use_saved_aggregation)
#        geo = aggregate_units(GEO(), dataset_name='GEO',
#                              aggregate_powerplant_units=False)
#        opsd = aggregate_units(OPSD(), dataset_name='OPSD',
#                               use_saved_aggregation=use_saved_aggregation)
#        wri = aggregate_units(WRI(), dataset_name='WRI',
#                              use_saved_aggregation=use_saved_aggregation)
#        if include_WEPP:
#            wepp = aggregate_units(WEPP(), dataset_name='WEPP',
#                                   use_saved_aggregation=use_saved_aggregation)
#    else:
#        carma = CARMA()
#        entsoe = ENTSOE()
#        geo = GEO()
#        opsd = OPSD()
#        wri = WRI()
#        if include_WEPP:
#            wepp = WEPP()
#    ese = ESE()
#    if include_WEPP:
#        stats = lookup(df=[carma, entsoe, ese, geo, opsd, wepp, wri],
#                       keys=['CARMA', 'ENTSO-E', 'ESE', 'GEO', 'OPSD', 'WEPP',
#                             'WRI'],
#                       exclude=exclude, by='Fueltype')/1000
#    else:
#        stats = lookup(df=[carma, entsoe, ese, geo, opsd, wri],
#                       keys=['CARMA', 'ENTSO-E', 'ESE', 'GEO', 'OPSD', 'WRI'],
#                       exclude=exclude, by='Fueltype')/1000
#    stats_reduced = lookup(df, by='Fueltype')/1000
#    # Presettings for the plots
#    with sns.axes_style(axes_style):
#        font = {'size': 12}
#        plt.rc('font', **font)
#        fig, ax = plt.subplots(nrows=1, ncols=2, sharex=False, sharey=True,
#                               figsize=figsize)
#        # 1st Plot with single datasets on the left side.
#        stats.plot.bar(ax=ax[0], stacked=False, legend=True, colormap='Accent')
#        ax[0].set_ylabel('Installed Capacity [$GW$]')
#        ax[0].set_title('Capacities of Single Databases')
#        ax[0].set_facecolor('#d9d9d9')                 # gray background
#        ax[0].set_axisbelow(True)                      # put grid behind bars
#        ax[0].grid(color='white', linestyle='dotted')  # adds white dotted grid
#        # 2nd Plot with reduced dataset
#        stats_reduced.plot.bar(ax=ax[1], stacked=False, colormap='jet')
#        ax[1].xaxis.label.set_visible(False)
#        ax[1].set_title('Capacities of Matched Dataset')
#        ax[1].set_facecolor('#d9d9d9')
#        ax[1].set_axisbelow(True)
#        ax[1].grid(color='white', linestyle='dotted')
#        fig.tight_layout()
#    return fig, ax
#
#
#def comparison_1dim(dfs=None, keys=None, by='Country', include_WEPP=True,
#                    include_VRE=False, year=2016, how='hbar',
#                    axes_style='whitegrid',
#                    exclude=['Geothermal', 'Solar', 'Wind', 'Battery'],
#                    **kwargs):
#    """
#    Plots a horizontal bar chart with capacity on x-axis, ``by`` on y-axis.
#
#    Parameters
#    ----------
#    by : string, defines how to group data
#        Allowed values: 'Country' or 'Fueltype'
#
#    """
#    if dfs is None and keys is None:
#
#        dfs = [df for df in gather_comparison_data(include_WEPP=include_WEPP,
#                                                   include_VRE=include_VRE,
#                                                   year=year)
#               if df is not None]
#        if include_WEPP:
#            keys = ['Matched dataset w/ WEPP', 'Matched dataset w/o WEPP',
#                    'WEPP only', 'Statistics ENTSO-E SO&AF']
#        else:
#            keys = ['Matched dataset w/o WEPP', 'Statistics ENTSO-E SO&AF']
#
#    stats = lookup(df=dfs, keys=keys, by=by, exclude=exclude)/1000
#
#    font = {'size': 12}
#    plt.rc('font', **font)
#    if how == 'hbar':
#        with sns.axes_style(axes_style):
#            figsize = kwargs.get('figsize', (7, 6))
#            fig, ax = plt.subplots(figsize=figsize)
#            stats.plot.barh(ax=ax, stacked=False)  # , colormap='jet')
#            ax.set_xlabel('Installed Capacity [GW]')
#            ax.yaxis.label.set_visible(False)
#            ax.set_facecolor('lightgrey')               # gray background
#            ax.set_axisbelow(True)                      # grid behind bars
#            ax.grid(color='white', linestyle='dotted')  # white dotted grid
#            ax.legend(loc='best')
#            ax.invert_yaxis()
#            return fig, ax
#    if how == 'scatter':
#        # Required for seaborn scatter plot
#        stats.loc[:, by] = stats.index.astype(str)
#        if len(stats.columns)-1 >= 3:
#            g = sns.PairGrid(stats, hue=by, palette='Set2')
#            g.map_lower(plt.scatter)
#            g.add_legend()
#        else:
#            g = sns.pairplot(stats, diag_kind='kde', hue=by, palette='Set2',
#                             x_vars=stats.columns[0], y_vars=stats.columns[1])
#        for i in range(0, len(g.axes)):
#            for j in range(0, len(g.axes[0])):
#                g.axes[i, j].set(xscale='log', yscale='log',
#                                 xlim=(1, 200), ylim=(1, 200))
#                if i != j:  # plot 45 degree identity line
#                    g.axes[i, j].plot([1, 200], [1, 200], 'k-',
#                                      alpha=0.25, zorder=0)
#                if i < j:
#                    g.axes[i, j].remove()
#                if i == j and i != 3:
#                    g.axes[i, j].annotate('Identical', (7, 10))
#        figsize = kwargs.get('figsize', None)
#        if figsize is not None:
#            g.fig.set_size_inches(figsize)
#        return g.fig, g.axes
#
#
#def gather_comparison_data(include_WEPP=True, include_VRE=False, **kwargs):
#    from .data import WEPP, Capacity_stats
#    from .collection import (
#            matched_data,
#            Carma_ENTSOE_ESE_GEO_GPD_IWPDCY_OPSD_WEPP_matched_reduced_VRE,
#            Carma_ENTSOE_ESE_GEO_GPD_IWPDCY_OPSD_matched_reduced_VRE,
#            Carma_ENTSOE_ESE_GEO_GPD_IWPDCY_OPSD_WEPP_matched_reduced)
#
#    yr = kwargs.get('year', 2016)
#    s = "Fueltype!='Solar' and Fueltype!='Wind' and Fueltype!='Geothermal'"
#    queryexpr = kwargs.get('queryexpr', s)
#    # 1+2: WEPP itself + Reduced w/ WEPP
#    s = '(YearCommissioned<={:04d}) or (YearCommissioned!=YearCommissioned)'
#    if include_WEPP:
#        wepp = WEPP()
#        wepp.query(s.format(yr), inplace=True)
#
#        if include_VRE:
#            red_w_wepp =\
#                Carma_ENTSOE_ESE_GEO_GPD_IWPDCY_OPSD_WEPP_matched_reduced_VRE()
#        else:
#            red_w_wepp =\
#                Carma_ENTSOE_ESE_GEO_GPD_IWPDCY_OPSD_WEPP_matched_reduced()
#            red_w_wepp.query(queryexpr, inplace=True)
#            wepp.query(queryexpr, inplace=True)
#        red_w_wepp.query(s.format(yr), inplace=True)
#    else:
#        wepp = None
#        red_w_wepp = None
#    # 3: Reduced w/o WEPP
#    if include_VRE:
#        red_wo_wepp =\
#            Carma_ENTSOE_ESE_GEO_GPD_IWPDCY_OPSD_matched_reduced_VRE()
#    else:
#        red_wo_wepp = matched_data()
#        # red_wo_wepp = Carma_ENTSOE_ESE_GEO_GPD_IWPDCY_OPSD_matched_reduced()
#        red_wo_wepp.query(queryexpr, inplace=True)
#    red_wo_wepp.query(s.format(yr), inplace=True)
#    # 4: Statistics
#    statistics = Capacity_stats(year=yr)
#    statistics.Fueltype.replace({'Mixed fuel types': 'Other'}, inplace=True)
#    if not include_VRE:
#        statistics.query(queryexpr, inplace=True)
#    return red_w_wepp, red_wo_wepp, wepp, statistics
#
#
#def matchcount_stats(df):
#    """
#    Plots the number of matches against the number of involved databases,
#    across all databases.
#    """
#    df = df.copy().iloc[:, -8:]
#    df.loc[:, 'MatchCount'] = df.notnull().sum(axis=1)
#    df.groupby(['MatchCount']).size().plot.bar()
#    return
#
#
#def gather_nrows_ncols(x, orientation='landscape'):
#    """
#    Derives [nrows, ncols] based on x plots, so that a subplot looks nicely.
#
#    Parameters
#    ----------
#    x : int, Number of subplots between [0, 42]
#    """
#    def calc(n, m):
#        if n <= 0:
#            n = 1
#        if m <= 0:
#            m = 1
#        while (n*m < x):
#            m += 1
#        return n, m
#
#    if not isinstance(x, int):
#        raise ValueError('An integer needs to be passed to this function.')
#    elif x <= 0:
#        raise ValueError('The given number of subplots is less or equal zero.')
#    elif x > 42:
#        raise ValueError("Are you sure that you want to put more than 42 "
#                         "subplots in one diagram?\n You better don't, it "
#                         "looks squeezed. Otherwise adapt the code.")
#    k = math.sqrt(x)
#    if k.is_integer():
#        return [int(k), int(k), 0]  # Square format!
#    else:
#        k = int(math.floor(k))
#        # Solution 1
#        n, m = calc(k, k+1)
#        sol1 = {'n': n, 'm': m, 'dif': (m*n) - x}
#        # Solution 2:
#        n, m = calc(k-1, k+1)
#        sol2 = {'n': n, 'm': m, 'dif': (m*n) - x}
#        if (((sol1['dif'] <= sol2['dif']) & (sol1['n'] >= 2)) |
#                (x in [7, 13, 14])):
#            n, m = [sol1['n'], sol1['m']]
#        else:
#            n, m = [sol2['n'], sol2['m']]
#        remainder = m*n - x
#        if orientation == 'landscape':
#            return n, m, remainder
#        elif orientation == 'portrait':
#            return m, n, remainder
#        else:
#            raise ValueError('Wrong `orientation` given!')
#
#
#
#
#orderdedfuels = ['Hydro',  # 'Solar', 'Wind',
#                 'Nuclear', 'Hard Coal', 'Lignite',
#                 'Oil', 'Natural Gas', 'Other']



# -*- coding: utf-8 -*-
# Copyright 2016-2018 Fabian Hofmann (FIAS), Jonas Hoersch (KIT, IAI) and
# Fabian Gotzens (FZJ, IEK-STE)

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Processed datasets of merged and/or adjusted data
"""
from __future__ import print_function

from .core import _data_out, get_config
from .utils import (set_uncommon_fueltypes_to_other, parmap,
                    to_dict_if_string, projectID_to_dict, set_column_name,
                    parse_if_not_stored)
from .heuristics import (extend_by_non_matched, extend_by_VRE)
from .cleaning import aggregate_units
from .matching import combine_multiple_datasets, reduce_matched_dataframe

import pandas as pd
import os
import logging
logger = logging.getLogger(__name__)


def collect(datasets, update=False, use_saved_aggregation=True,
            use_saved_matches=True, reduced=True,
            custom_config={}, config=None, **dukeargs):
    """
    Return the collection for a given list of datasets in matched or
    reduced form.

    Parameters
    ----------
    datasets : list or str
        list containing the dataset identifiers as str, or single str
    update : bool
        Do an horizontal update (True) or read from the cache file (False)
    use_saved_aggregation : bool
        Aggregate units based on cached aggregation group files (True)
        or to do an vertical update (False)
    use_saved_matches : bool
        Match datasets based on cached matched pair files (True)
        or to do an horizontal matching (False)
    reduced : bool
        Switch as to return the reduced (True) or matched (False) dataset.
    custom_config : dict
        Updates to the data_config dict from data module
    **dukeargs : keyword-args for duke
    """

    from . import data
    if config is None:
        config = get_config()

    def df_by_name(name):
        conf = config[name]

        get_df = getattr(data, name)
        df = get_df(config=config, **conf.get('read_kwargs', {}))
        if not conf.get('aggregated_units', False):
            return aggregate_units(df,
                                   use_saved_aggregation=use_saved_aggregation,
                                   dataset_name=name,
                                   config=config)
        else:
            return df.assign(projectID=df.projectID.map(lambda x: [x]))

    # Deal with the case that only one dataset is requested
    if isinstance(datasets, str):
        return df_by_name(datasets)

    datasets = sorted(datasets)
    logger.info('Collect combined dataset for {}'.format(', '.join(datasets)))
    outfn_matched = _data_out('Matched_{}.csv'
                              .format('_'.join(map(str.upper, datasets))),
                              config=config)
    outfn_reduced = _data_out('Matched_{}_reduced.csv'
                              .format('_'.join(map(str.upper, datasets))),
                              config=config)

    if not update and not os.path.exists(outfn_reduced
                                         if reduced else outfn_matched):
        logger.warning("Forcing update since the cache file is missing")
        update = True
        use_saved_aggregation = True

    if update:
        dfs = parmap(df_by_name, datasets)
        matched = combine_multiple_datasets(
                dfs, datasets, use_saved_matches=use_saved_matches,
                config=config, **dukeargs)
        (matched.assign(projectID=lambda df: df.projectID.astype(str))
                .to_csv(outfn_matched, index_label='id'))

        reduced_df = reduce_matched_dataframe(matched, config=config)
        reduced_df.to_csv(outfn_reduced, index_label='id')

        return reduced_df if reduced else matched
    else:
        if reduced:
            df = pd.read_csv(outfn_reduced, index_col=0)
        else:
            df = pd.read_csv(outfn_matched, index_col=0, header=[0, 1],
                             low_memory=False)
        return df.pipe(projectID_to_dict)


def Collection(**kwargs):
    return collect(**kwargs)


def matched_data(config=None,
                 stored=True,
                 update=False,
                 update_all=False,
                 from_url=False,
                 extend_by_vres=False,
                 extendby_kwargs={'use_saved_aggregation': True},
                 subsume_uncommon_fueltypes=False,
                 **collection_kwargs):
    """
    Return the full matched dataset including all data sources listed in
    config.yaml/matching_sources. The combined data is additionally extended
    by non-matched entries of sources given in
    config.yaml/fully_inculded_souces.


    Parameters
    ----------
    stored : Boolean, default True
            Whether to use the stored matched_data.csv file in data/out/default
            If False, the matched data is taken from collect() and
            extended afterwards. To update the whole matching, please set
            stored=False and update=True.
    update : Boolean, default False
            Whether to rerun the matching process. Overrides stored to False
            if True.
    update_all : Boolean, default False
            Whether to rerun the matching process and aggregation process.
            Overrides stored to False if True.
    from_url: Boolean, default False
            Whether to parse and store the already build data from the repo
            website.
    config : Dict, default None
            Define a configuration varying from the setting in config.yaml.
            Relevant keywords are 'matching_sources', 'fully_included_sources'.
    extend_by_vres : Boolean, default False
            Whether extend the dataset by variable renewable energy sources
            given by powerplantmatching.data.OPSD_VRE()
    extendby_kwargs : Dict, default {'use_saved_aggregation': True}
            Dict of keywordarguments passed to powerplatnmatchting.
            heuristics.extend_by_non_matched
    subsume_uncommon_fueltypes : Boolean, default False
            Whether to replace uncommon fueltype specification by 'Other'
    **collection_kwargs : kwargs
            Arguments passed to powerplantmatching.collection.Collection.
            Typical arguments are update, use_saved_aggregation,
            use_saved_matches.

    """
    if config is None:
        config = get_config()

    if update_all:
        collection_kwargs['use_saved_aggregation'] = False
        collection_kwargs['use_saved_matches'] = False
        update = True
    if update:
        stored = False

    collection_kwargs.setdefault('update', update)

    if collection_kwargs.get('reduced', True):
        fn = _data_out('matched_data_red.csv')
        header = 0
    else:
        fn = _data_out('matched_data.csv')
        header = [0, 1]

    if from_url:
        fn = _data_out('matched_data_red.csv')
        url = config['matched_data_url']
        logger.info(f'Retrieving data from {url}')
        df = (pd.read_csv(url, index_col=0)
                .pipe(projectID_to_dict)
                .pipe(set_column_name, 'Matched Data'))
        logger.info(f'Store data at {fn}')
        df.to_csv(fn)
        return df

    if stored and os.path.exists(fn):
        df = (pd.read_csv(fn, index_col=0, header=header)
                .pipe(projectID_to_dict)
                .pipe(set_column_name, 'Matched Data'))
        if extend_by_vres:
                return df.pipe(extend_by_VRE, config=config,
                               base_year=config['opsd_vres_base_year'])
        return df

    matching_sources = [list(to_dict_if_string(a))[0] for a in
                                  config['matching_sources']]
    matched = collect(matching_sources, **collection_kwargs)

    if isinstance(config['fully_included_sources'], list):
        for source in config['fully_included_sources']:
            source = to_dict_if_string(source)
            name, = list(source)
            extendby_kwargs.update({'query': source[name]})
            matched = extend_by_non_matched(matched, name, config=config,
                                            **extendby_kwargs)

    # Drop matches between only low reliability-data, this is necessary since
    # a lot of those are decommissioned, however some countries only appear in
    # GEO and CARMA
    allowed_countries = config['CARMA_GEO_countries']
    if matched.columns.nlevels > 1:
        other = set(matching_sources) - set(['CARMA', 'GEO'])
        matched = (matched[~matched.projectID[other].isna().all(1) |
                           matched.Country.GEO.isin(allowed_countries) |
                           matched.Country.CARMA.isin(allowed_countries)]
                   .reset_index(drop=True))
        if config['remove_missing_coords']:
            matched = (matched[matched.lat.notnull().any(1)]
                       .reset_index(drop=True))
    else:
        matched = (matched[matched.projectID.apply(lambda x: sorted(x.keys())
                           not in [['CARMA', 'GEO']]) |
                           matched.Country.isin(allowed_countries)]
                   .reset_index(drop=True))
        if config['remove_missing_coords']:
            matched = matched[matched.lat.notnull()].reset_index(drop=True)
    matched.to_csv(fn, index_label='id', encoding='utf-8')

    if extend_by_vres:
        matched = extend_by_VRE(matched, config=config,
                                base_year=config['opsd_vres_base_year'])

    if subsume_uncommon_fueltypes:
        matched = set_uncommon_fueltypes_to_other(matched)
    return matched.pipe(set_column_name, 'Matched Data')





# -*- coding: utf-8 -*-
# Copyright 2016-2018 Fabian Hofmann (FIAS), Jonas Hoersch (KIT, IAI) and
# Fabian Gotzens (FZJ, IEK-STE)

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Functions for linking and combining different datasets
"""

from __future__ import absolute_import, print_function

from .core import get_config, _data_out, get_obj_if_Acc
from .utils import read_csv_if_string, parmap, get_name
from .duke import duke
from .cleaning import clean_technology

import os.path
import pandas as pd
import numpy as np
from itertools import combinations
import logging
logger = logging.getLogger(__name__)


def best_matches(links):
    """
    Subsequent to duke() with singlematch=True. Returns reduced list of
    matches on the base of the highest score for each duplicated entry.

    Parameters
    ----------
    links : pd.DataFrame
        Links as returned by duke
    """
    labels = links.columns.difference({'scores'})
    return (links
            .groupby(links.iloc[:, 1], as_index=False, sort=False)
            .apply(lambda x: x.loc[x.scores.idxmax(), labels]))


def compare_two_datasets(dfs, labels, use_saved_matches=False,
                         country_wise=True, config=None, **dukeargs):
    """
    Duke-based horizontal match of two databases. Returns the matched
    dataframe including only the matched entries in a multi-indexed
    pandas.Dataframe. Compares all properties of the given columns
    ['Name','Fueltype', 'Technology', 'Country',
    'Capacity','lat', 'lon'] in order to determine the same
    powerplant in different two datasets. The match is in one-to-one
    mode, that is every entry of the initial databases has maximally
    one link in order to obtain unique entries in the resulting
    dataframe.  Attention: When aborting this command, the duke
    process will still continue in the background, wait until the
    process is finished before restarting.

    Parameters
    ----------
    dfs : list of pandas.Dataframe or strings
        dataframes or csv-files to use for the matching
    labels : list of strings
        Names of the databases for the resulting dataframe


    """
    if config is None:
        config = get_config()

    dfs = list(map(read_csv_if_string, dfs))
    if not ('singlematch' in dukeargs):
        dukeargs['singlematch'] = True
    saving_path = _data_out('matches/matches_{}_{}.csv'
                            .format(*np.sort(labels)), config=config)
    if use_saved_matches:
        if os.path.exists(saving_path):
            logger.info('Reading saved matches for dfs {} and {}'
                        .format(*labels))
            return pd.read_csv(saving_path, index_col=0)
        else:
            logger.warning("Non-existing saved matches for dataset '{}', '{}'"
                           " continuing by matching again".format(*labels))

    def country_link(dfs, country):
        #country_selector for both dataframes
        sel_country_b = [df['Country'] == country for df in dfs]
        #only append if country appears in both dataframse
        if all(sel.any() for sel in sel_country_b):
            return duke([df[sel] for df, sel in zip(dfs, sel_country_b)],
                         labels, **dukeargs)
        else:
            return pd.DataFrame()

    if country_wise:
        countries = config['target_countries']
        links = pd.concat([country_link(dfs, c) for c in countries])
    else:
        links = duke(dfs, labels=labels, **dukeargs)
    matches = best_matches(links)
    matches.to_csv(saving_path)
    return matches


def cross_matches(sets_of_pairs, labels=None):
    """
    Combines multiple sets of pairs and returns one consistent
    dataframe. Identifiers of two datasets can appear in one row even
    though they did not match directly but indirectly through a
    connecting identifier of another database.

    Parameters
    ----------
    sets_of_pairs : list
        list of pd.Dataframe's containing only the matches (without
        scores), obtained from the linkfile (duke() and
        best_matches())
    labels : list of strings
        list of names of the databases, used for specifying the order
        of the output

    """
    m_all = sets_of_pairs
    if labels is None:
        labels = np.unique([x.columns for x in m_all])
    matches = pd.DataFrame(columns=labels)
    for i in labels:
        base = [m.set_index(i) for m in m_all if i in m]
        match_base = pd.concat(base, axis=1).reset_index()
        matches = pd.concat([matches, match_base], sort=True)

    matches = matches.drop_duplicates().reset_index(drop=True)
    for i in labels:
        matches = pd.concat([
            matches.groupby(i, as_index=False, sort=False)
                   .apply(lambda x: x.loc[x.isnull().sum(axis=1).idxmin()]),
            matches[matches[i].isnull()]
        ]).reset_index(drop=True)
    return (matches
            .assign(length=matches.notna().sum(axis=1))
            .sort_values(by='length', ascending=False)
            .reset_index(drop=True)
            .drop('length', axis=1)
            .reindex(columns=labels))


def link_multiple_datasets(datasets, labels, use_saved_matches=False,
                           config=None, **dukeargs):
    """
    Duke-based horizontal match of multiple databases. Returns the
    matching indices of the datasets. Compares all properties of the
    given columns ['Name','Fueltype', 'Technology', 'Country',
    'Capacity','lat', 'lon'] in order to determine the same
    powerplant in different datasets. The match is in one-to-one mode,
    that is every entry of the initial databases has maximally one
    link to the other database.  This leads to unique entries in the
    resulting dataframe.

    Parameters
    ----------
    datasets : list of pandas.Dataframe or strings
        dataframes or csv-files to use for the matching
    labels : list of strings
        Names of the databases in alphabetical order and corresponding
        order to the datasets
    """
    if config is None:
        config = get_config()

    dfs = list(map(read_csv_if_string, datasets))
    labels = [get_name(df) for df in dfs]

    combs = list(combinations(range(len(labels)), 2))

    def comp_dfs(dfs_lbs):
        logger.info('Comparing {0} with {1}'.format(*dfs_lbs[2:]))
        return compare_two_datasets(dfs_lbs[:2], dfs_lbs[2:],
                                    use_saved_matches=use_saved_matches,
                                    config=config, **dukeargs)

    mapargs = [[dfs[c], dfs[d], labels[c], labels[d]] for c, d in combs]
    all_matches = parmap(comp_dfs, mapargs)

    return cross_matches(all_matches, labels=labels)


def combine_multiple_datasets(datasets, labels=None, use_saved_matches=False,
                              config=None, **dukeargs):
    """
    Duke-based horizontal match of multiple databases. Returns the
    matched dataframe including only the matched entries in a
    multi-indexed pandas.Dataframe. Compares all properties of the
    given columns ['Name','Fueltype', 'Technology', 'Country',
    'Capacity','lat', 'lon'] in order to determine the same
    powerplant in different datasets. The match is in one-to-one mode,
    that is every entry of the initial databases has maximally one
    link to the other database.  This leads to unique entries in the
    resulting dataframe.

    Parameters
    ----------
    datasets : list of pandas.Dataframe or strings
        dataframes or csv-files to use for the matching
    labels : list of strings
        Names of the databases in alphabetical order and corresponding
        order to the datasets
    """
    if config is None:
        config = get_config()

    def combined_dataframe(cross_matches, datasets, config):
        """
        Use this function to create a matched dataframe on base of the
        cross matches and a list of the databases. Always order the
        database alphabetically.

        Parameters
        ----------
        cross_matches : pandas.Dataframe of the matching indexes of
            the databases, created with
            powerplant_collection.cross_matches()
        datasets : list of pandas.Dataframes or csv-files in the same
            order as in cross_matches
        """
        datasets = list(map(read_csv_if_string, datasets))
        for i, data in enumerate(datasets):
            datasets[i] = (data
                           .reindex(cross_matches.iloc[:, i])
                           .reset_index(drop=True))
        return (pd.concat(datasets, axis=1,
                          keys=cross_matches.columns.tolist())
                .reorder_levels([1, 0], axis=1)
                .reindex(columns=config['target_columns'], level=0)
                .reset_index(drop=True))
    crossmatches = link_multiple_datasets(datasets, labels,
                                          use_saved_matches=use_saved_matches,
                                          config=config, **dukeargs)
    return (combined_dataframe(crossmatches, datasets, config)
            .reindex(columns=config['target_columns'], level=0))


def reduce_matched_dataframe(df, show_orig_names=False, config=None):
    """
    Reduce a matched dataframe to a unique set of columns. For each entry
    take the value of the most reliable data source included in that match.

    Parameters
    ----------
    df : pandas.Dataframe
        MultiIndex dataframe with the matched powerplants, as obtained from
        combined_dataframe() or match_multiple_datasets()
    """
    df = get_obj_if_Acc(df)

    if config is None:
        config = get_config()


    # define which databases are present and get their reliability_score
    sources = df.columns.levels[1]
    rel_scores = pd.Series({s: config[s]['reliability_score'] for s in sources})\
                   .sort_values(ascending=False)
    cols = config['target_columns']
    props_for_groups = {col : 'first'
                        for col in cols}
    props_for_groups.update({'YearCommisisoned': 'min',
                     'Retrofit': 'max',
                     'projectID': lambda x: dict(x.droplevel(0).dropna()),
                     'eic_code': 'unique'})
    props_for_groups = pd.Series(props_for_groups)[cols].to_dict()

    #set low priority on Fueltype 'Other'
    #turn it since aggregating only possible for axis=0
    sdf = df.replace({'Fueltype' : {'Other': np.nan}})\
            .stack(1).reindex(rel_scores.index, level=1)\
            .groupby(level=0)\
            .agg(props_for_groups)\
            .replace({'Fueltype' : {np.nan: 'Other'}})


    if show_orig_names:
        sdf = sdf.assign(**dict(df.Name))
    return sdf.pipe(clean_technology).reset_index(drop=True)



from __future__ import absolute_import



from setuptools import setup, find_packages
from codecs import open


with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='powerplantmatching',
    version='0.4.2',
    author='Fabian Hofmann (FIAS), Jonas Hoersch (KIT), Fabian Gotzens (FZ Jülich)',
    author_email='hofmann@fias.uni-frankfurt.de',
    description='Toolset for generating and managing Power Plant Data',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/FRESNA/powerplantmatching',
    license='GPLv3',
#    packages=find_packages(include='matching_analysis'),
    packages=['powerplantmatching'],
    include_package_data=True,
    install_requires=['numpy','scipy','pandas>=0.23.0','networkx>=1.10',
                      'pycountry', 'xlrd', 'seaborn', 'pyyaml', 'requests',
                      'matplotlib', 'geopy', 'xlrd', 'entsoe-py'],
    classifiers=[
        'Programming Language :: Python :: 3',
#        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
    ])



# -*- coding: utf-8 -*-
# Copyright 2016-2018 Fabian Hofmann (FIAS), Jonas Hoersch (KIT, IAI) and
# Fabian Gotzens (FZJ, IEK-STE)

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Functions for vertically cleaning a dataset.
"""
from __future__ import absolute_import, print_function

from .core import get_config, _data_out, get_obj_if_Acc
from .utils import get_name, set_column_name
from .duke import duke

import os
import numpy as np
import pandas as pd
import networkx as nx
import logging
logger = logging.getLogger(__name__)


def clean_powerplantname(df):
    """
    Cleans the column "Name" of the database by deleting very frequent
    words, numericals and nonalphanumerical characters of the
    column. Returns a reduced dataframe with nonempty Name-column.

    Parameters
    ----------
    df : pandas.Dataframe
        dataframe to be cleaned

    """
    df = get_obj_if_Acc(df)
    df = df[df.Name.notnull()]
    name = df.Name.replace(regex=True, value=' ',
                           to_replace=['-', '/', ',', '\(', '\)', '\[', '\]',
                                       '"', '_', '\+', '[0-9]'])

    common_words = pd.Series(sum(name.str.split(), [])).value_counts()
    cw = list(common_words[common_words >= 20].index)

    pattern = [('(?i)(^|\s)'+x+'(?=\s|$)')
               for x in (cw +
                         ['[a-z]', 'I', 'II', 'III', 'IV', 'V', 'VI', 'VII',
                          'VIII', 'IX', 'X', 'XI', 'Grupo', 'parque', 'eolico',
                          'gas', 'biomasa', 'COGENERACION', 'gt', 'unnamed',
                          'tratamiento de purines', 'planta', 'de', 'la',
                          'station', 'power', 'storage', 'plant', 'stage',
                          'pumped', 'project', 'dt', 'gud', 'hkw', 'kbr',
                          'Kernkraft', 'Kernkraftwerk', 'kwg', 'krb', 'ohu',
                          'gkn', 'Gemeinschaftskernkraftwerk', 'kki', 'kkp',
                          'kle', 'wkw', 'rwe', 'bis', 'nordsee', 'ostsee',
                          'dampfturbinenanlage', 'ikw', 'kw', 'kohlekraftwerk',
                          'raffineriekraftwerk', 'Kraftwerke'])]
    name = (name
            .replace(regex=True, to_replace=pattern, value=' ')
            .replace('\s+', ' ', regex=True)
            .replace('"', '', regex=True)
            .str.strip()
            .str.capitalize())

    return (df
            .assign(Name=name)
            .loc[lambda x: x.Name != '']
            .sort_values('Name')
            .reset_index(drop=True))


def gather_fueltype_info(df, search_col=['Name', 'Technology']):
    """
    Parses in search_col columns for distinct coal specifications, e.g.
    'lignite', and passes this information to the 'Fueltype' column.

    Parameter
    ---------
    search_col : list, default is ['Name', 'Technology']
        Specify the columns to be parsed
    """
    fueltype = pd.Series(df['Fueltype'])

    for i in search_col:
        found_b = df[i].dropna().str.contains('(?i)lignite|brown')
        fueltype.loc[found_b.reindex(fueltype.index,
                                     fill_value=False)] = 'Lignite'
    fueltype.replace({'Coal': 'Hard Coal'}, inplace=True)

    return df.assign(Fueltype=fueltype)


def gather_technology_info(df, search_col=['Name', 'Fueltype'],
                           config=None):
    """
    Parses in search_col columns for distinct technology specifications, e.g.
    'Run-of-River', and passes this information to the 'Technology' column.

    Parameter
    ---------
    search_col : list, default is ['Name', 'Fueltype']
        Specify the columns to be parsed
    config : dict, default None
        Add custom specific configuration,
        e.g. powerplantmatching.config.get_config(target_countries='Italy'),
        defaults to powerplantmatching.config.get_config()

    """
    if config is None:
        config = get_config()

    technology = (df['Technology'].dropna()
                  if 'Technology' in df
                  else pd.Series())

    pattern = '|'.join(('(?i)'+x) for x in config['target_technologies'])
    for i in search_col:
        found = (df[i].dropna()
                 .str.findall(pattern)
                 .loc[lambda s: s.str.len() > 0]
                 .str.join(sep=', '))

        exists_i = technology.index.intersection(found.index)
        if len(exists_i) > 0:
            technology.loc[exists_i] = (technology.loc[exists_i].str
                                        .cat(found.loc[exists_i], sep=', '))

        new_i = found.index.difference(technology.index)
        technology = technology.append(found[new_i])

    return df.assign(Technology=technology)


def gather_set_info(df, search_col=['Name', 'Fueltype', 'Technology']):
    """
    Parses in search_col columns for distinct set specifications, e.g.
    'Store', and passes this information to the 'Set' column.

    Parameter
    ---------
    search_col : list, default is ['Name', 'Fueltype', 'Technology']
        Specify the columns to be parsed
    config : dict, default None
        Add custom specific configuration,
        e.g. powerplantmatching.config.get_config(target_countries='Italy'),
        defaults to powerplantmatching.config.get_config()

    """
    Set = (df['Set'].copy()
           if 'Set' in df
           else pd.Series(index=df.index))

    pattern = '|'.join(['heizkraftwerk', 'hkw', 'chp', 'bhkw', 'cogeneration',
                        'power and heat', 'heat and power'])
    for i in search_col:
        isCHP_b = df[i].dropna().str.contains(pattern, case=False)\
                    .reindex(df.index).fillna(False)
        Set.loc[isCHP_b] = 'CHP'

    pattern = '|'.join(['battery', 'storage'])
    for i in search_col:
        isStore_b = df[i].dropna().str.contains(pattern, case=False) \
                    .reindex(df.index).fillna(False)
        Set.loc[isStore_b] = 'Store'

    df = df.assign(Set=Set)
    df.loc[:, 'Set'].fillna('PP', inplace=True)
    return df


def clean_technology(df, generalize_hydros=False):
    """
    Clean the 'Technology' by condensing down the value into one claim. This
    procedure might reduce the scope of information, however is crucial for
    comparing different data sources.

    Parameter
    ---------
    search_col : list, default is ['Name', 'Fueltype', 'Technology']
        Specify the columns to be parsed
    config : dict, default None
        Add custom specific configuration,
        e.g. powerplantmatching.config.get_config(target_countries='Italy'),
        defaults to powerplantmatching.config.get_config()

    """
    tech = df['Technology'].dropna()
    if len(tech) == 0:
        return df
    tech = tech.replace(
            {' and ': ', ', ' Power Plant': '', 'Battery': ''}, regex=True)
    if generalize_hydros:
        tech[tech.str.contains('pump', case=False)] = 'Pumped Storage'
        tech[tech.str.contains('reservoir|lake', case=False)] = 'Reservoir'
        tech[tech.str.contains('run-of-river|weir|water', case=False)] =\
            'Run-Of-River'
        tech[tech.str.contains('dam', case=False)] = 'Reservoir'
    tech = tech.replace({'Gas turbine': 'OCGT'})
    tech[tech.str.contains('combined cycle|combustion', case=False)] = 'CCGT'
    tech[tech.str.contains('steam turbine|critical thermal', case=False)] =\
        'Steam Turbine'
    tech[tech.str.contains('ocgt|open cycle', case=False)] = 'OCGT'
    tech = (tech.str.title()
                .str.split(', ')
                .apply(lambda x: ', '.join(i.strip() for i in np.unique(x))))
    tech = tech.replace({'Ccgt': 'CCGT', 'Ocgt': 'OCGT'}, regex=True)
    return df.assign(Technology=tech)


def cliques(df, dataduplicates):
    """
    Locate cliques of units which are determined to belong to the same
    powerplant.  Return the same dataframe with an additional column
    "grouped" which indicates the group that the powerplant is
    belonging to.

    Parameters
    ----------
    df : pandas.Dataframe or string
        dataframe or csv-file which should be analysed
    dataduplicates : pandas.Dataframe or string
        dataframe or name of the csv-linkfile which determines the
        link within one dataset
    """
#    df = read_csv_if_string(df)
    G = nx.DiGraph()
    G.add_nodes_from(df.index)
    G.add_edges_from((r.one, r.two) for r in dataduplicates.itertuples())
    H = G.to_undirected(reciprocal=True)

    grouped = pd.Series(np.nan, index=df.index)
    for i, inds in enumerate(nx.algorithms.clique.find_cliques(H)):
        grouped.loc[inds] = i

    return df.assign(grouped=grouped)

def aggregate_units(df, dataset_name=None,
                    pre_clean_name=True,
                    save_aggregation=True,
                    country_wise=True,
                    use_saved_aggregation=False,
                    config=None):
    """
    Vertical cleaning of the database. Cleans the "Name"-column, sums
    up the capacity of powerplant units which are determined to belong
    to the same plant.

    Parameters
    ----------
    df : pandas.Dataframe or string
        Dataframe or name to use for the resulting database
    dataset_name : str, default None
        Specify the name of your df, required if use_saved_aggregation is set
        to True.
    pre_clean_name : Boolean, default True
        Whether to clean the 'Name'-column before aggregating.
    use_saved_aggregation : bool (default False):
        Whether to use the automaticly saved aggregation file, which
        is stored in data/out/default/aggregations/aggregation_groups_XX.csv
        with XX being the name for the dataset. This saves time if you
        want to have aggregated powerplants without running the
        aggregation algorithm again
    """
    df = get_obj_if_Acc(df)

    if config is None:
        config = get_config()

    weighted_cols = [col for col in ['Efficiency', 'Duration']
                     if col in config['target_columns']]
    df = (df.assign(**{col: df[col] * df.Capacity for col in weighted_cols})
            .assign(lat=df.lat.astype(float),
                    lon=df.lon.astype(float))
            .assign(**{col: df[col].astype(str) for col in
                       ['Name', 'Country', 'Fueltype',
                        'Technology', 'Set', 'File']
                       if col in config['target_columns']}))

    def mode(x):
        return x.mode(dropna=False).at[0]

    props_for_groups = pd.Series({
                'Name': mode,
                'Country': mode,
                'Fueltype': mode,
                'Technology': mode,
                'Set': mode,
                'File': mode,
                'Capacity': 'sum',
                'lat': 'mean',
                'lon': 'mean',
                'YearCommissioned': 'min',
                'Retrofit': 'max',
                'projectID': list,
                'eic_code': set,
                'Duration': 'sum',  # note this is weighted sum
                'Volume_Mm3': 'sum',
                'DamHeight_m': 'sum',
                'Efficiency': 'mean'  # note this is weighted mean
                })[config['target_columns']].to_dict()

    dataset_name = get_name(df) if dataset_name is None else dataset_name

    if pre_clean_name:
        df = clean_powerplantname(df)

    logger.info("Aggregating blocks to entire units in '{}'."
                .format(dataset_name))

    path_name = _data_out('aggregations/aggregation_groups_{}.csv'
                          .format(dataset_name), config=config)


    if use_saved_aggregation & save_aggregation:
        if os.path.exists(path_name):
            logger.info("Reading saved aggregation groups for dataset '{}'."
                        .format(dataset_name))
            groups = (pd.read_csv(path_name, header=None, index_col=0)
                        .reindex(index=df.index))
            df = df.assign(grouped=groups.values)
        else:
            logger.info("Non-existing saved aggregation groups for dataset"
                           " '{0}', continuing by aggregating again"
                           .format(dataset_name))
            if 'grouped' in df:
                df.drop('grouped', axis=1, inplace=True)

    if 'grouped' not in df:
        if country_wise:
            duplicates = pd.concat([duke(df.query('Country == @c'))
                                    for c in df.Country.unique()])
        else:
            duplicates = duke(df)
        df = cliques(df, duplicates)
        if save_aggregation:
            df.grouped.to_csv(path_name, header=False)

    df = df.groupby('grouped').agg(props_for_groups)
    df = df.replace('nan', np.nan)

    if 'eic_code' in df:
        df = df.assign(eic_code = df['eic_code'].apply(list))

    df = (df
          .assign(**{col: df[col].div(df['Capacity'])
                  for col in weighted_cols})
          .reset_index(drop=True)
          .pipe(clean_powerplantname)
          .reindex(columns=config['target_columns'])
          .pipe(set_column_name, dataset_name))
    return df



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jul 16 15:47:46 2019

@author: fabian
"""

import os
import logging
import pandas.api.extensions

# for the writable data directory (i.e. the one where new data goes), follow
# the XDG guidelines found at
# https://standards.freedesktop.org/basedir-spec/basedir-spec-latest.html
_writable_dir = os.path.join(os.path.expanduser('~'), '.local', 'share')
_data_dir = os.path.join(os.environ.get("XDG_DATA_HOME",
                            os.environ.get("APPDATA", _writable_dir)),
                                'powerplantmatching')
#data file configuration
package_config = {
        'custom_config': os.path.join(os.path.expanduser('~'),
                                      '.powerplantmatching_config.yaml'),
        'data_dir': _data_dir,
        'repo_data_dir': os.path.join(os.path.dirname(__file__), 'package_data'),
        'downloaders': {}}



def _package_data(fn):
    return os.path.join(package_config['repo_data_dir'], fn)


def _data_in(fn):
    return os.path.join(package_config['data_dir'], 'data', 'in', fn)


def _data_out(fn, config=None):
    if config is None:
        return os.path.join(package_config['data_dir'], 'data', 'out',
                            'default', fn)
    else:
        return os.path.join(package_config['data_dir'], 'data', 'out',
                            config['hash'], fn)

del _data_dir
del _writable_dir


if not os.path.exists(_data_in('.')):
    os.makedirs(_data_in('.'))

# Logging: General Settings
logger = logging.getLogger(__name__)
logging.basicConfig(level=20)
logger.setLevel('INFO')
# Logging: File
logFormatter = logging.Formatter("%(asctime)s [%(threadName)-12.12s] " +
                                 "[%(levelname)-5.5s]  %(message)s")
fileHandler = logging.FileHandler(
        os.path.join(package_config['data_dir'], 'PPM.log'))
fileHandler.setFormatter(logFormatter)
logger.addHandler(fileHandler)
# logger.info('Initialization complete.')

del logFormatter
del fileHandler



def get_config(filename=None, **overrides):
    from hashlib import sha1
    from base64 import encodestring
    from six.moves import cPickle
    import yaml
    from logging import info

    if filename is None:
        custom_config = package_config['custom_config']
        if os.path.exists(custom_config):
            filename = custom_config
        else:
            filename = _package_data('config.yaml')

    with open(filename) as f:
        config = yaml.load(f, Loader=yaml.FullLoader)
        config.update(overrides)

        sha1digest = sha1(cPickle.dumps(overrides)).digest()
        if len(dict(**overrides)) == 0:
            config['hash'] = 'default'
        else:
            config['hash'] = encodestring(sha1digest)\
                             .decode('ascii')[2:12]
    if not os.path.isdir(_data_out('.', config=config)):
        os.makedirs(os.path.abspath(_data_out('.', config=config)))
        os.makedirs(os.path.abspath(_data_out('matches', config=config)))
        os.makedirs(os.path.abspath(_data_out('aggregations', config=config)))
        info('Outputs for this configuration will be saved under {}'
                    .format(os.path.abspath(
                            _data_out('.', config=config))))
        with open(_data_out('config.yaml', config=config), 'w') as file:
            yaml.dump(config, file, default_flow_style=False)
    return config


def get_obj_if_Acc(obj):
    from .accessor import PowerPlantAccessor
    if isinstance(obj, PowerPlantAccessor):
        return obj._obj
    else:
        return obj

